/*
 *	Sherlock Filter Engine --- dumping the parsed filter
 *
 *	(c) 2002 Robert Spalek <robert@ucw.cz>
 */

#include "lib/lib.h"
#include "lib/fastbuf.h"
#include "lib/printf.h"
#include "lib/conf.h"
#include "filter/filter.h"
#include "filter/parse.tab.h"

static void
dump_spaces(struct fastbuf *b, uns level)
{
	uns i;
	for (i=0; i<level; i++)
		bputc(b, '\t');
}

static void
dump_declarations(struct fastbuf *b, struct filter_declaration *decl)
{
	for (; decl; decl=decl->next)
	{
		byte *type;
		ASSERT(decl->var.cat == F_LVC_USER);
		ASSERT(!decl->var.ro);
		switch (decl->var.type)
		{
			case F_ET_INT: type = "int"; break;
			case F_ET_STRING: type = "string"; break;
			case F_ET_REGEXP: type = "regexp"; break;
			default: ASSERT(0);
		}
		bputs(b, xprintf("%s %s_%d;\n", type, decl->var.name, decl->nr));
	}
}

static void
dump_variable_name(struct fastbuf *b, struct filter_lvalue *lval)
{
	struct cfitem *cfi;
	switch (lval->cat)
	{
		case F_LVC_RAW:
			if (lval->undef)
				bputs(b, lval->v.bind_name);
			else
				bputs(b, lval->v.bind->name);
			break;
		case F_LVC_ATTR:
			bputs(b, xprintf("attr[\"%c\"]", lval->v.name));
			break;
		case F_LVC_CONF:
			cfi = lval->v.cfg;
			if (cfi)
			{
				while (cfi->type != CT_SECTION && cfi->type != CT_INCOMPLETE_SECTION)
					cfi--;
				bputs(b, xprintf("conf[%s.%s]", cfi->name, lval->v.cfg->name));
			}
			else
				bputs(b, "conf[?.?]");
			break;
		case F_LVC_USER:
			bputs(b, xprintf("%s_%d", lval->v.decl->var.name, lval->v.decl->nr));
			break;
		default:
			ASSERT(0);
	}
}

static void
dump_expression(struct fastbuf *b, struct filter_expr *e)
{
	int i;
	switch (e->cat)
	{
		case F_EC_CONST:
			switch (e->type)
			{
				case F_ET_INT: bputs(b, xprintf("%d", e->o.i)); break;
				case F_ET_STRING: bputs(b, xprintf("\"%s\"", e->o.s)); break;
				case F_ET_REGEXP: bputs(b, "<regexp>"); break;
				default: ASSERT(0);
			}
			break;
		case F_EC_LVALUE:
			dump_variable_name(b, e->o.lv);
			break;
		case F_EC_UNOP:
			bputc(b, e->o.un.op);
			dump_expression(b, e->o.un.r);
			break;
		case F_EC_BINOP:
			bputs(b, " (");
			dump_expression(b, e->o.bin.l);
			bputc(b, ' ');
			bputc(b, e->o.bin.op);
			bputc(b, ' ');
			dump_expression(b, e->o.bin.r);
			bputs(b, ") ");
			break;
		case F_EC_FUNC:
			bputs(b, e->o.func.func->name);
			bputc(b, '(');
			for (i=0; i<e->o.func.func->args; i++)
			{
				dump_expression(b, e->o.func.a[i]);
				if (i < e->o.func.func->args-1)
					bputs(b, ", ");
			}
			bputc(b, ')');
			break;
		default:
			ASSERT(0);
	}
}

static void
dump_condition(struct fastbuf *b, struct filter_cond *c)
{
	byte *msg = NULL;
	switch (c->cat)
	{
		case F_CC_CONST:
			switch (c->o.i)
			{
				case TRUE: bputs(b, "true"); break;
				case FALSE: bputs(b, "false"); break;
				case UNDEFINED: bputs(b, "undefined"); break;
				default: ASSERT(0);
			}
			break;
		case F_CC_EXPR:
			if (c->o.expr.l)
			{
				bputs(b, " (");
				dump_expression(b, c->o.expr.l);
			}
			bputc(b, ' ');
			if (!c->o.expr.neg)
			{
				switch (c->o.expr.op)
				{
					case EREG: msg = "=~"; break;
					case EPAT: msg = "=*"; break;
					case EQ: msg = "=="; break;
					case LT: msg = "<"; break;
					case GT: msg = ">"; break;
					default: ASSERT(0);
				}
			}
			else
			{
				switch (c->o.expr.op)
				{
					case EREG: msg = "!~"; break;
					case EPAT: msg = "!*"; break;
					case EQ: msg = "!="; break;
					case LT: msg = ">="; break;
					case GT: msg = "<="; break;
					default: ASSERT(0);
				}
			}
			bputs(b, msg);
			if (c->o.expr.icase)
				bputc(b, msg[strlen(msg) - 1]);
			bputc(b, ' ');
			dump_expression(b, c->o.expr.r);
			if (c->o.expr.l)
				bputs(b, ") ");
			break;
		case F_CC_DEFCOND:
			bputs(b, "defined(");
			dump_condition(b, c->o.neg);
			bputs(b, ") ");
			break;
		case F_CC_DEFEXPR:
			bputs(b, "defined(");
			dump_expression(b, c->o.defexpr);
			bputs(b, ") ");
			break;
		case F_CC_UNOP:
			bputc(b, '!');
			dump_condition(b, c->o.neg);
			break;
		case F_CC_BINOP:
			bputs(b, " (");
			dump_condition(b, c->o.bin.l);
			switch (c->o.bin.op)
			{
				case AND: bputs(b, " && "); break;
				case OR: bputs(b, " || "); break;
				case EQ: bputs(b, " == "); break;
				case NE: bputs(b, " != "); break;
				default: ASSERT(0);
			}
			dump_condition(b, c->o.bin.r);
			bputs(b, ") ");
			break;
		default:
			ASSERT(0);
	}
}

static void dump_commands(struct fastbuf *b, struct filter_cmd *cmd, uns level);

static void
dump_switch_hash(struct fastbuf *b, struct filter_hash_table *ht, byte *op, uns level)
{
	uns i;
	struct filter_hash_record *hr;
	if (!ht)
		return;
	dump_spaces(b, level);
	bputs(b, xprintf("# hash-table for the operator %s\n", op));
	for (i=0; i<ht->size; i++)
		for (hr=ht->h[i]; hr; hr=hr->next)
		{
			dump_spaces(b, level);
			bputs(b, "case ");
			bputs(b, op);
			bputs(b, " \"");
			bputs(b, hr->string);
			bputs(b, "\":\n");
			dump_commands(b, hr->cmd, level+1);
		}
}

static void
dump_switch_cases(struct fastbuf *b, struct filter_case *fc, uns level)
{
	if (fc)
	{
		dump_spaces(b, level);
		bputs(b, "# sequence of SWITCH cases\n");
	}
	for (; fc; fc=fc->next)
	{
		dump_spaces(b, level);
		bputs(b, "case ");
		dump_condition(b, fc->cond);
		bputs(b, ":\n");
		dump_commands(b, fc->positive, level+1);
	}
}

static void
dump_commands(struct fastbuf *b, struct filter_cmd *cmd, uns level)
{
	for (; cmd; cmd=cmd->next)
	{
		byte *msg = NULL;
		if (!cmd->op)
			continue;
		dump_spaces(b, level);
		switch (cmd->op)
		{
			case LOG1:
				switch (cmd->c.print.level)
				{
					case L_DEBUG: msg = "debug"; break;
					case L_INFO_R: msg = "log"; break;
					case L_WARN_R: msg = "warning"; break;
					case L_ERROR_R: msg = "error"; break;
					default: ASSERT(0);
				}
				bputs(b, msg);
				bputc(b, ' ');
				dump_expression(b, cmd->c.print.expr);
				bputs(b, ";\n");
				break;
			case ACCEPT:
				msg = "accept";
				/* Fall-thru.  */
			case REJECT:
				if (!msg)
					msg = "reject";
				bputs(b, msg);
				if (cmd->c.print.expr)
				{
					bputc(b, ' ');
					dump_expression(b, cmd->c.print.expr);
				}
				bputs(b, ";\n");
				break;
			case ADD:
				msg = "add ";
				/* Fall-thru.  */
			case '=':
				if (msg)
					bputs(b, msg);
				dump_variable_name(b, cmd->c.set.lv);
				bputs(b, " = ");
				dump_expression(b, cmd->c.set.expr);
				bputs(b, ";\n");
				break;
			case DELETE:
				bputs(b, "delete ");
				dump_variable_name(b, cmd->c.set.lv);
				bputs(b, ";\n");
				break;
			case IF:
				bputs(b, "if ");
				dump_condition(b, cmd->c.cond.cond);
				bputs(b, " {\n");
				dump_commands(b, cmd->c.cond.positive, level+1);
				dump_spaces(b, level);
				bputc(b, '}');
				if (cmd->c.cond.negative)
				{
					bputs(b, " else {\n");
					dump_commands(b, cmd->c.cond.negative, level+1);
					dump_spaces(b, level);
					bputc(b, '}');
				}
				if (cmd->c.cond.undefined)
				{
					bputs(b, " undef {\n");
					dump_commands(b, cmd->c.cond.undefined, level+1);
					dump_spaces(b, level);
					bputc(b, '}');
				}
				bputc(b, '\n');
				break;
			case SWITCH:
				bputs(b, "switch ");
				dump_expression(b, cmd->c.swit.expr);
				bputs(b, " {\n");
				dump_switch_hash(b, cmd->c.swit.cmp, "==", level+1);
				dump_switch_hash(b, cmd->c.swit.icmp, "===", level+1);
				dump_switch_cases(b, cmd->c.swit.cases, level+1);
				dump_spaces(b, level);
				bputc(b, '}');
				if (cmd->c.swit.negative)
				{
					bputs(b, " else {\n");
					dump_commands(b, cmd->c.swit.negative, level+1);
					dump_spaces(b, level);
					bputc(b, '}');
				}
				if (cmd->c.swit.undefined)
				{
					bputs(b, " undef {\n");
					dump_commands(b, cmd->c.swit.undefined, level+1);
					dump_spaces(b, level);
					bputc(b, '}');
				}
				bputc(b, '\n');
				break;
			default:
				ASSERT(0);
		}
	}
}

void
filter_dump(struct fastbuf *b, struct filter *f)
{
	bputs(b, "# Global variables\n");
	dump_declarations(b, f->decl);
	bputs(b, "\n# Context variables\n");
	dump_declarations(b, f->deleted_decl);
	bputs(b, "\n# Program\n");
	dump_commands(b, f->body, 0);
}
