/*
 *	Sherlock Gatherer Daemon -- The Expirer
 *
 *	(c) 1997--2002 Martin Mares <mj@ucw.cz>
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "lib/conf.h"
#include "lib/url.h"
#include "lib/db.h"
#include "lib/fastbuf.h"
#include "gather/daemon/gatherd.h"

#include <time.h>
#include <setjmp.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>

/*** Globals ***/

sh_time_t now, cycle_start, cycle_last;

static uns do_not_expire;
static uns one_pass;

/*** Configuration ***/

static int min_revalidate_age = 0x7fffffff;
static uns revalidate_cycle;
static int queue_discard_age = 0x7fffffff;
static int error_discard_age = 0x7fffffff;
static int robot_expire_age = 0x7fffffff;
static byte *tmp_queue_name = "db/queue.tmp";
static uns qk_expire_age = ~0U;
static byte *stamp_name = "db/expire-stamp";
static int trace;
static int queue_postpone = 0x7fffffff;
static uns hist_num_boxes = 0;
static uns hist_box_width = 86400;

static struct cfitem expire_config[] = {
  { "Expire",		CT_SECTION,	NULL },
  { "Trace",		CT_INT,		&trace },
  { "MinRevalidateAge",	CT_INT,		&min_revalidate_age },
  { "RevalidateCycle",	CT_INT,		&revalidate_cycle },
  { "QueueDiscardAge",	CT_INT,		&queue_discard_age },
  { "ErrorDiscardAge",	CT_INT,		&error_discard_age },
  { "TmpQueueFile",	CT_STRING,	&tmp_queue_name },
  { "QKeyExpireAge",	CT_INT,		&qk_expire_age },
  { "StampFile",	CT_STRING,	&stamp_name },
  { "RobotExpireAge",	CT_INT,		&robot_expire_age },
  { "QueuePostpone",	CT_INT,		&queue_postpone },
  { "HistNumBoxes",	CT_INT,		&hist_num_boxes },
  { "HistBoxWidth",	CT_INT,		&hist_box_width },
  { NULL,		CT_STOP,	NULL }
};

#define TRACE(msg) do { if (trace) log(L_DEBUG, msg); } while(0)

static void
stamp_read(void)
{
  FILE *f;

  if (f = fopen(stamp_name, "r"))
    {
      if (fscanf(f, "%d%d", &cycle_start, &cycle_last) != 2)
	log(L_ERROR, "%s: syntax error, ignoring the stamp", stamp_name);
      fclose(f);
    }
  if (!cycle_last)
    { 
      log(L_INFO, "Restarting expiration cycle");
      cycle_start = cycle_last = now;
    }
}

static void
stamp_write(void)
{
  FILE *f;

  if (!(f = fopen(stamp_name, "w")))
    die("fopen(%s): %m", stamp_name);
  fprintf(f, "%d %d\n", (int) cycle_start, (int) now);
  fclose(f);
}

/*** Here Comes The Judge ***/

static struct rfilter_data rfdata;
static int j_low, j_high;

#define DENOM 4096			/* Common denominator of all our fractions */

static int
judge_hash(byte *name)
{
  uns h = strlen(name);

  while (*name)
    h = h*37 + *name++;
  return h & (DENOM-1);
}

static int
judge_filter(byte *url)
{
  byte *msg;

  rfdata.url = url;
  ref_filter_undo();
  msg = ref_filter(&rfdata);
  if (msg)
    {
      if (trace > 1)
	log(L_DEBUG, "%s: %s", url, msg);
      return 0;
    }
  return 3;
}

static void
judge_init(void)
{
  if (now - cycle_last >= revalidate_cycle)
    {
      j_low = 0;
      j_high = DENOM;
    }
  else
    {
      j_low = (int)((float)(cycle_last - cycle_start) * DENOM / revalidate_cycle) % DENOM;
      j_high = (int)((float)(now - cycle_start) * DENOM / revalidate_cycle) % DENOM;
    }
  log(L_DEBUG, "Revalidation cycle: %d--%d of %d", j_low, j_high, DENOM);
}

/* Return values: 0=filtered out, 1=expire, 2=rescan, 3=ok */
static int
judge(byte *url, struct urlrec *ur)
{
  int age, verdict;
  byte *msg = "OK";
  static int warned_time_skew;

  verdict = judge_filter(url);
  if (verdict != 3)
    return verdict;
  age = now - ur->access;
  if (age < 0)
    {
      if (!warned_time_skew++)
	log(L_ERROR, "Clock skew detected");
      age = 0;
    }
  if (do_not_expire)
    msg = "not expired";
  else if (ur->flags & URF_QUEUED)
    {
      if (age > queue_discard_age)
	{
	  msg = "expiring from queue";
	  verdict = 1;
	}
      else
	msg = "already queued";
    }
  else if (ur->oid >= OID_FIRST_ERROR)
    {
      if (age > error_discard_age)
	{
	  msg = "expiring error";
	  verdict = 1;
	}
      else
	msg = "error marker";
    }
  else if (age > min_revalidate_age)
    {
      int j = judge_hash(url);
      if ((j_low <= j_high) ?
	  (j >= j_low && j < j_high) :
	  (j >= j_low || j < j_high))
	{
	  msg = "revalidate";
	  verdict = 2;
	}
      else
	msg = "waiting for revalidation";
    }
  if (trace > 1)
    log(L_DEBUG, "%s: %s", url, msg);
  return verdict;
}

/* The ID Array */

struct id {
  oid_t oid;				/* Original OID */
  oid_t new;				/* New OID after renumbering (0=not seen, OID_UNDEFINED if deleted) */
};

#define ID_ARRAY_SIZE 65536

struct id_array {
  struct id_array *next;
  uns count;
  struct id id[ID_ARRAY_SIZE];
};

static struct id_array **id_arrays, *id_first_arr, *id_this_arr;
static uns id_num_arrays, id_count;

#define ID(i) (id_arrays[(i)/ID_ARRAY_SIZE]->id[(i)%ID_ARRAY_SIZE])

static void
ids_init(void)
{
  id_num_arrays = 1;
  id_first_arr = id_this_arr = xmalloc(sizeof(struct id_array));
  id_this_arr->next = NULL;
  id_this_arr->count = 0;
}

static inline void
ids_add(oid_t oid, oid_t new)
{
  struct id *id;

  if (id_this_arr->count >= ID_ARRAY_SIZE)
    {
      id_this_arr = id_this_arr->next = xmalloc(sizeof(struct id_array));
      id_this_arr->next = NULL;
      id_this_arr->count = 0;
      id_num_arrays++;
    }
  id = &id_this_arr->id[id_this_arr->count++];
  id->oid = oid;
  id->new = new;
}

static void
ids_close(void)
{
  uns i = 0;

  id_arrays = xmalloc(sizeof(struct id_array *) * id_num_arrays);
  while (id_first_arr)
    {
      id_arrays[i++] = id_first_arr;
      id_count += id_first_arr->count;
      id_first_arr = id_first_arr->next;
    }
  if (trace)
    log(L_DEBUG, "Found %d object ID's (internally: %d blocks)", id_count, id_num_arrays);
}

static void
ids_sort(void)
{
  /*
   *  Unfortunately we cannot use qsort() here as we use a rather
   *  non-standard representation of our data by disconnected
   *  arrays. We do a simple QuickSort instead with a couple
   *  of standard tricks a`la Sedgewick.
   */
  struct stk { int l, r; } stack[32];
  int l, r, left, right, m;
  uns sp = 0;
  oid_t pivot;
  struct id tmp;
#define SWAP(x,y) do { tmp=ID(x); ID(x)=ID(y); ID(y)=tmp; } while(0)

  left = 0;
  right = id_count - 1;
  if (left >= right)
    return;
  for(;;)
    {
      l = left;
      r = right;
      m = (l+r)/2;
      if (ID(l).oid > ID(m).oid)
	SWAP(l,m);
      if (ID(m).oid > ID(r).oid)
	{
	  SWAP(m,r);
	  if (ID(l).oid > ID(m).oid)
	    SWAP(l,m);
	}
      pivot = ID(m).oid;
      do
	{
	  while (ID(l).oid < pivot)
	    l++;
	  while (ID(r).oid > pivot)
	    r--;
	  if (l < r)
	    {
	      SWAP(l,r);
	      l++;
	      r--;
	    }
	  else if (l == r)
	    {
	      l++;
	      r--;
	    }
	}
      while (l <= r);
      if (left < r && l < right)
	{
	  /* Both partitions ok => push the larger one */
	  if ((r - left) > (right - l))
	    {
	      stack[sp].l = left;
	      stack[sp].r = r;
	      left = l;
	    }
	  else
	    {
	      stack[sp].l = l;
	      stack[sp].r = right;
	      right = r;
	    }
	  sp++;
	}
      else if (left < r)
	{
	  /* Left partition OK, right undersize */
	  right = r;
	}
      else if (l < right)
	{
	  /* Right partition OK, left undersize */
	  left = l;
	}
      else
	{
	  /* Both partitions undersize => pop */
	  if (!sp)
	    break;
	  sp--;
	  left = stack[sp].l;
	  right = stack[sp].r;
	}
    }
#undef SWAP
}

static int
ids_lookup(oid_t oid)
{
  int l = 0;
  int r = id_count-1;
  int m;

  while (l <= r)
    {
      m = (l+r)/2;
      if (ID(m).oid == oid)
	return m;
      if (ID(m).oid < oid)
	l = m+1;
      else
	r = m-1;
    }
  die("ID %08x lost", oid);
}

static void
ids_debug(byte *msg)
{
  uns i;
  if (trace < 3)
    return;
  log(L_DEBUG, "%s", msg);
  for (i=0; i<id_count; i++)
    {
      ASSERT(!i || ID(i).oid > ID(i-1).oid);
      fprintf(stderr, "\t%08x %08x\n", ID(i).oid, ID(i).new);
    }
}

static inline oid_t
oid_translate(oid_t oid)
{
  if (!one_pass && oid < OID_FIRST_ERROR)
    {
      int i = ids_lookup(oid);
      oid = ID(i).new;
    }
  return oid;
}

static inline void
oid_delete(oid_t oid)
{
  if (one_pass)
    {
      if (oid < OID_FIRST_ERROR)
	ids_add(oid, OID_UNDEFINED);
    }
  else
    ASSERT(oid >= OID_FIRST_ERROR);
}

/*** Initial scan of URL database ***/

static void
scan_urls(void)
{
  byte url[MAX_URL_SIZE];
  struct urlrec ur;
  uns total=0, del=0;

  TRACE("Prescanning URL database");
  urldb_open();
  urldb_rewind();
  while (urldb_get_next(url, &ur))
    {
      if (ur.oid < OID_FIRST_ERROR)
	{
	  int verdict = judge(url, &ur);
	  total++;
	  if (verdict >= 2)
	    ids_add(ur.oid, 0);
	  else
	    {
	      del++;
	      ids_add(ur.oid, OID_UNDEFINED);
	    }
	}
    }
  urldb_close();
  log(L_INFO, "URL prescan: total %d buckets, %d to delete", total, del);
}

/*** Expiration of robot records ***/

static uns rob_total, rob_del, rob_filter;

static int
judge_robot(struct qhost *h)
{
  struct url ur;
  byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE];
  int e, age;
  byte *msg = "OK";

  if (h->robot_id == OID_UNDEFINED)
    return 3;
  ur.protocol = url_proto_names[h->protocol];
  ur.protoid = h->protocol;
  ur.port = h->port;
  ur.user = NULL;
  ur.host = h->name;
  ur.rest = "/robots.txt";
  if ((e = url_pack(&ur, buf1)) || (e = url_enescape(buf1, buf2)))
    {
      log(L_ERROR, "Error creating robots.txt URL: %s", url_error(e));
      return 0;
    }
  e = judge_filter(buf2);
  if (e != 3)
    return e;
  age = now - h->robot_time;
  if (do_not_expire)
    msg = "not expired";
  else if (age > robot_expire_age)
    {
      e = 1;
      msg = "aged out";
      h->rec_err_count = 0;
    }
  if (trace > 1)
    log(L_DEBUG, "%s: %s", buf2, msg);
  return e;
}

static void
scan_robots_host(struct qhost *h)
{
  if (!h->protocol)			/* Free list entry */
    return;
  if (h->robot_id < OID_FIRST_ERROR)
    {
      rob_total++;
      if (judge_robot(h) >= 2)
	ids_add(h->robot_id, 0);
      else
	{
	  rob_del++;
	  ids_add(h->robot_id, OID_UNDEFINED);
	}
    }
}

static void
scan_robots(void)
{
  TRACE("Prescanning robot records");
  walk_hosts(scan_robots_host);
  log(L_INFO, "Robot prescan: total %d buckets, %d to delete", rob_total, rob_del);
}

static void
chew_robots_host(struct qhost *h)
{
  int verdict;

  if (!h->protocol)			/* Free list entry */
    return;
  if (h->robot_id == OID_UNDEFINED)	/* No robots.txt */
    return;
  verdict = judge_robot(h);
  h->robot_id = oid_translate(h->robot_id);
  rob_total++;
  if (verdict < 2)
    {
      oid_delete(h->robot_id);
      h->robot_id = OID_UNDEFINED;
      h->robot_time = now;
      if (verdict)
	rob_del++;
      else
	rob_filter++;
    }
}

static void
chew_robots(void)
{
  TRACE("Expiring robot records");
  rob_total = rob_del = rob_filter = 0;
  walk_hosts(chew_robots_host);
  log(L_INFO, "Robots: %d total, %d expired, %d filtered out", rob_total, rob_del, rob_filter);
}

/*** Expiration of queue keys ***/

static int qk_low, qk_high;
static uns qk_seen, qk_zero, qk_expired;

static void
chew_qkeys_host(struct qhost *h)
{
  int q;

  qk_seen++;
  if (!h->qkey)
    {
      qk_zero++;
      return;
    }
  q = judge_hash(h->name);
  if ((qk_low <= qk_high) ?
      (q >= qk_low && q < qk_high) :
      (q >= qk_low || q < qk_high))
    {
      qk_expired++;
      h->qkey = 0;
    }
}

static void
chew_qkeys(void)
{
  if (qk_expire_age == ~0U || do_not_expire)
    return;
  TRACE("Expiring queue keys");
  if (now - cycle_last >= qk_expire_age)
    {
      qk_low = 0;
      qk_high = DENOM;
    }
  else
    {
      qk_low = (int)((float)(cycle_last - cycle_start) * DENOM / qk_expire_age) % DENOM;
      qk_high = (int)((float)(now - cycle_start) * DENOM / qk_expire_age) % DENOM;
    }
  DBG("Expiring %d--%d/%d", qk_low, qk_high, DENOM);
  walk_hosts(chew_qkeys_host);
  log(L_INFO, "Queue keys: %d total with %d unset, %d expired [%d--%d of %d]",
      qk_seen, qk_zero, qk_expired, qk_low, qk_high, DENOM);
}

/*** Shaking down buckets and reconstructing MD5 database ***/

static uns shake_rover;
static uns shake_total, shake_deleted_old, shake_deleted_new, shake_errors;
static uns md5_sums, md5_dups;

static inline void
shake_md5(oid_t new, byte *buck, byte *end)
{
  while (buck < end)
    {
      if (*buck == 'C')
	{
	  if (buck + 34 > end)
	    log(L_ERROR, "Object %08x: Malformed MD5 attribute", new);
	  else if (md5db_exists(buck+1))
	    md5_dups++;
	  else
	    {
	      md5db_store(buck+1, new);
	      md5_sums++;
	    }
	  return;
	}
      while (buck < end && *buck++ != '\n')
	;
    }
}

static int
shake_kibitz(struct obuck_header *old, oid_t new, byte *buck)
{
  shake_total++;
  if (new == OBUCK_OID_DELETED)
    {
      shake_deleted_old++;
      return 0;
    }
  while (shake_rover < id_count && ID(shake_rover).oid < old->oid)
    {
      log(L_ERROR, "Bucket %08x missing", ID(shake_rover).oid);
      ID(shake_rover).new = OID_UNDEFINED;
      shake_rover++;
      shake_errors++;
    }
  if (shake_rover >= id_count || ID(shake_rover).oid > old->oid)
    {
      log(L_ERROR, "Bucket %08x unreferenced, deleting it", old->oid);
      shake_errors++;
      return 0;
    }
  if (ID(shake_rover).new == OID_UNDEFINED)
    {
      shake_rover++;
      shake_deleted_new++;
      return 0;
    }
  else
    {
      ID(shake_rover).new = new;
      shake_rover++;
      shake_md5(new, buck, buck+old->length);
      return 1;
    }
}

static void
shake(void)
{
  TRACE("Rebuilding MD5 database");
  if (unlink(md5db_name) < 0 && errno != ENOENT)
    die("Cannot unlink old MD5 database: %m");
  md5db_open();

  TRACE("Shaking down buckets");
  obuck_shakedown(shake_kibitz);

  while (shake_rover < id_count)
    {
      log(L_ERROR, "Bucket %08x missing", ID(shake_rover).oid);
      ID(shake_rover).new = OID_UNDEFINED;
      shake_rover++;
      shake_errors++;
    }

  log(L_INFO, "Buckets: %d total, %d old + %d new deleted, %d bad",
      shake_total, shake_deleted_old, shake_deleted_new, shake_errors);
  log(L_INFO, "MD5: %d sums + %d duplicates", md5_sums, md5_dups);
  md5db_close();
}

/*** Quick deletion without bucket file shakedown ***/

static void
quick_delete(void)
{
  uns i;
  uns count=0;

  TRACE("Deleting unused buckets");
  for (i=0; i<id_count; i++)
    {
      struct id *id = &ID(i);
      if (id->new == OID_UNDEFINED)
	{
	  DBG("Deleting %08x", id->oid);
	  obuck_delete(id->oid);
	  count++;
	}
      else
	id->new = id->oid;
    }
  log(L_INFO, "Deleted %d buckets", count);
}

/*** Remaking the queue ***/

#define SORT_KEY struct qs_key
#define SORT_PREFIX(x) qs_##x
#define SORT_PRESORT
#define SORT_DELETE_INPUT 1
#define SORT_INPUT_FB
#define SORT_OUTPUT_FB

struct qs_key {
  u32 weight;
  addr_int_t host;
  uns len;
};

static inline int
qs_compare(struct qs_key *a, struct qs_key *b)
{
  if (a->host < b->host)
    return -1;
  if (a->host > b->host)
    return 1;
  if (a->weight > b->weight)
    return -1;
  if (a->weight < b->weight)
    return 1;
  return 0;
}

static inline int
qs_fetch_key(struct fastbuf *f, struct qs_key *k)
{
  k->weight = bgetl(f);
  k->host = bgeta(f);
  k->len = bgetw(f);
  return (k->weight != 0xffffffff);
}

static inline void
qs_copy_data(struct fastbuf *src, struct fastbuf *dest, struct qs_key *k)
{
  bputl(dest, k->weight);
  bputa(dest, k->host);
  bputw(dest, k->len);
  bbcopy(src, dest, k->len+8);
}

static inline byte *
qs_fetch_item(struct fastbuf *f, struct qs_key *k, byte *limit)
{
  byte *buf = (byte *)(k+1);
  if (buf + k->len + 8 > limit)
    return NULL;
  breadb(f, buf, k->len+8);
  return buf + k->len+8;
}

static inline void
qs_store_item(struct fastbuf *dest, struct qs_key *k)
{
  bputl(dest, k->weight);
  bputa(dest, k->host);
  bputw(dest, k->len);
  bwrite(dest, (byte *)(k+1), k->len+8);
}

#include "lib/sorter.h"

#define QL_MAGIC 0x430121dd

static void
rebuild_queue(struct fastbuf *b)
{
  u32 weight, maxobj, section;
  uns len, count=0, inactive=0, postponed=0;
  struct qhost *h;
  byte name[MAX_URL_SIZE];

  TRACE("Sorting queue items");
  bflush(b);
  bsetpos(b, 0);
  b = qs_sort(b);

  TRACE("Rebuilding queue");
  while ((weight = bgetl(b)) != 0xffffffff)
    {
      h = (struct qhost *) bgeta(b);
      len = bgetw(b);
      ASSERT(len && len < MAX_URL_SIZE);
      breadb(b, name, len);
      name[len] = 0;
      section = bgetl(b);
      maxobj = bgetl(b);
      DBG("Q: h=%s n=%s w=%d s=%d m=%d", h->name, name, weight, section, maxobj);
      if (section != ~0U && ++h->obj_count[section] > maxobj)
	inactive++;
      else if (++h->ex_qh_qcount > (uns) queue_postpone)
	postponed++;
      else
	{
	  enqueue_item(h, name);
	  count++;
	}
    }
  bclose(b);
  log(L_INFO, "Queue: %d items, %d inactive, %d postponed", count, inactive, postponed);
}

/*** Chewing the URL database ***/

static uns url_stats[4][4];
static uns url_in[4], url_out[4], url_lost;

static inline uns
classify_ur(struct urlrec *ur)
{
  if (ur->flags & URF_QUEUED)
    {
      if (ur->oid == OID_UNDEFINED)
	return 0;			/* Type 0 = queued new */
      else
	return 1;			/* Type 1 = queued refresh */
    }
  else if (ur->oid >= OID_FIRST_ERROR)
    return 2;				/* Type 2 = error */
  else
    return 3;				/* Type 3 = normal document */
}

static void
chew_urls(void)
{
  struct sdbm_options so;
  struct sdbm *newdb;
  byte url[MAX_URL_SIZE];
  struct urlrec ur;
  byte newdb_name[strlen(urldb_name) + 5];
  struct qhost *h;
  struct fastbuf *qlist;
  uns *hist, hboxes, hbox, timestamp;
  int age;

  TRACE("Expiring URL database");

  /* Initialize the histogram */
  hboxes = hist_num_boxes+1;
  hist = alloca(sizeof(uns) * 4 * hboxes);
  bzero(hist, sizeof(uns) * 4 * hboxes);

  /* Create new URL database */
  sprintf(newdb_name, "%s.tmp", urldb_name);
  if (unlink(newdb_name) < 0 && errno != ENOENT)
    die("Cannot unlink old URL database: %m");
  /* Must match parameters in urldb_open() */
  so.name = newdb_name;
  so.flags = SDBM_CREAT | SDBM_WRITE;
  so.page_order = 12;
  so.cache_size = urldb_cache_size;
  so.key_size = -1;
  so.val_size = sizeof(struct urlrec);
  if (!(newdb = sdbm_open(&so)))
    die("Cannot open the URL database");

  /* Create the list of to-be-queued items */
  qlist = bopen(tmp_queue_name, O_CREAT|O_TRUNC|O_RDWR, 1048576);

  urldb_open();
  urldb_rewind();
  queue_reset();
  while (urldb_get_next(url, &ur))
    {
      int verdict = judge(url, &ur);
      uns type = classify_ur(&ur);
      url_in[type]++;
      ur.oid = oid_translate(ur.oid);
      DBG("%s: %08x %c %d%s", url, ur.oid, "FDRO"[verdict], now-ur.access, (ur.flags&URF_QUEUED) ? " [q]" : "");
      if (ur.oid == OID_UNDEFINED && verdict >= 2 && !(ur.flags & URF_QUEUED))
	{
	  log(L_ERROR, "Object %s lost", url);
	  url_lost++;
	  continue;
	}
      url_stats[type][verdict]++;

      /* Copy to new URLdb and adjust according to verdict */
      switch (verdict)
	{
	case 0:				/* discard */
	case 1:
	  oid_delete(ur.oid);
	  continue;
	case 2:				/* revalidate */
	  if (!(ur.flags & URF_QUEUED))
	    {
	      ur.access = now;
	      ur.flags |= URF_QUEUED;
	    }
	  break;
	default:			/* ok */
	  ;
	}
      if (!sdbm_store(newdb, url, strlen(url), (byte *) &ur, sizeof(ur)))
	log(L_ERROR, "Found and ignored duplicate URL: %s", url);
      type = classify_ur(&ur);
      url_out[type]++;

      /*
       *  The host's object count handling is going to be somewhat tricky:
       *  URLdb contains records even for overflowed URL's and we want to
       *  to include as many of them as possible without exceeding the limit.
       *  Hence we update object count here for already downloaded URL's
       *  (which must not be dropped) only and do the rest when rebuilding
       *  the queue.
       */
      h = find_host(rfdata.url_s.protoid, rfdata.url_s.host, rfdata.url_s.port);
      if (!h)
	{
	  log(L_WARN, "Creating host entry for %s", url);
	  h = new_host(rfdata.url_s.protoid, rfdata.url_s.host, rfdata.url_s.port);
	}
      if (ur.oid != OID_UNDEFINED)
	{
	  h->obj_count[rfdata.section]++;
	  if (h->obj_count[rfdata.section] == rfdata.section_hard_max+1)
	    log(L_WARN, "Host %s overfull (probably changed limits, don't worry)", rfdata.url_s.host);
	}

      /* Put to the queue if needed */
      timestamp = ur.access;
      if (timestamp > now)
	timestamp = now;
      age = now - timestamp;
      if (ur.flags & URF_QUEUED)
	{
	  u32 w = age;
	  if (ur.oid != OID_UNDEFINED)
	    w += 0x80000000;	/* Refreshes always have precedence */
	  bputl(qlist, w);
	  bputa(qlist, (addr_int_t) h);
	  bputw(qlist, strlen(rfdata.url_s.rest));
	  bputs(qlist, rfdata.url_s.rest);
	  bputl(qlist, (ur.oid == OID_UNDEFINED) ? rfdata.section : ~0U);
	  bputl(qlist, rfdata.section_soft_max);
	}

      /* Add to histogram */
      hbox = ((now - now % hist_box_width + hist_box_width) - timestamp) / hist_box_width;
      if (hbox > hist_num_boxes)
	hbox = hist_num_boxes;
      hist[4*hbox+type]++;
    }
  urldb_close();
  sdbm_close(newdb);
  if (rename(newdb_name, urldb_name))
    die("urldb rename: %m");
  log(L_INFO, "Total URL's: %d in -> %d out: %d expired, %d requeued, %d filtered out, %d lost",
      url_in[0] + url_in[1] + url_in[2] + url_in[3],
      url_out[0] + url_out[1] + url_out[2] + url_out[3],
      url_stats[0][1] + url_stats[1][1] + url_stats[2][1] + url_stats[3][1],
      url_stats[0][2] + url_stats[1][2] + url_stats[2][2] + url_stats[3][2],
      url_stats[0][0] + url_stats[1][0] + url_stats[2][0] + url_stats[3][0],
      url_lost);
  log(L_INFO, "Queued new URL's: %d in -> %d out: %d ok, %d expired, %d filtered out",
      url_in[0], url_out[0], url_stats[0][3], url_stats[0][1], url_stats[0][0]);
  log(L_INFO, "Queued refresh URL's: %d in -> %d out: %d ok, %d expired, %d filtered out",
      url_in[1], url_out[1], url_stats[1][3], url_stats[1][1], url_stats[1][0]);
  log(L_INFO, "Static URL's: %d in -> %d out: %d ok, %d requeued, %d expired, %d filtered out",
      url_in[3], url_out[3], url_stats[3][3], url_stats[3][2], url_stats[3][1], url_stats[3][0]);
  log(L_INFO, "Error URL's: %d in -> %d out: %d ok, %d expired, %d filtered out",
      url_in[2], url_out[2], url_stats[2][3], url_stats[2][1], url_stats[2][0]);

  if (hist_num_boxes)
    {
      byte hh[hboxes*48], *h = hh;
      for (hbox=0; hbox<hboxes; hbox++)
	h += sprintf(h, " %d/%d/%d/%d", hist[4*hbox], hist[4*hbox+1], hist[4*hbox+2], hist[4*hbox+3]);
      log(L_INFO, "Age histogram (new/rfsh/err/stat):%s", hh);
    }

  rebuild_queue(qlist);
}

/*** Main ***/

static void NONRET
usage(void)
{
  fprintf(stderr, "Usage: expire [<switches>]\n\n"
CF_USAGE
"-f\t\tOnly filter URL's and rebuild the queue, don't expire anything\n\
-q\t\tQuick mode, don't shake down buckets nor rebuild MD5db\n\
-v\t\tIncrease verbosity (a.k.a. Expire.Trace)\n\
");
  die("Invalid arguments.");
}

int
main(int argc, char **argv)
{
  int c;

  log_init(argv[0]);
  cf_register(expire_config);
  while ((c = cf_getopt(argc, argv, "fqv" CF_SHORT_OPTS, CF_NO_LONG_OPTS, NULL)) >= 0)
    switch (c)
      {
      case 'f':
	do_not_expire = 1;
	break;
      case 'q':
	one_pass = 1;
	break;
      case 'v':
	trace++;
	break;
      default:
	usage();
      }
  if (optind != argc)
    usage();

  log(L_INFO, "Starting expirer");
  now = time(NULL);
  gather_lock();
  queue_init();
  refs_init(1);
  obuck_init(1);
  stamp_read();

  ids_init();
  judge_init();
  if (one_pass)
    {
      chew_urls();
      chew_robots();
      chew_qkeys();
      ids_close();
      ids_sort();
      ids_debug("ID's to be deleted");
      quick_delete();
    }
  else
    {
      scan_urls();
      scan_robots();
      ids_close();
      ids_sort();
      ids_debug("Initial table of ID's");
      shake();
      ids_debug("ID's after shakedown");
      chew_urls();
      chew_robots();
      chew_qkeys();
    }

  stamp_write();
  obuck_cleanup();
  queue_cleanup();
  gather_unlock();
  log(L_INFO, "Expiration finished in %d seconds", (int)(time(NULL) - now));
  return 0;
}
