/*
 *	Sherlock Indexer -- Digestive Processes
 *
 *	(c) 2001--2002 Martin Mares <mj@ucw.cz>
 */

#include "lib/lib.h"
#include "lib/conf.h"
#include "lib/fastbuf.h"
#include "lib/pools.h"
#include "lib/url.h"
#include "lib/hashfunc.h"
#include "charset/unicode.h"
#include "indexer/indexer.h"
#include "indexer/lexicon.h"

#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

static struct fastbuf *cards, *card_attrs, *word_index, *string_index;
static uns card_id;
static byte *descr_stop;

#undef PROFILE_TSC
#include "lib/profile.h"
static prof_t pr_fetch, pr_strings, pr_words, pr_cards;

/*
 *  Configuration
 */

static uns word_buf_size = 65536;
static uns string_buf_size = 65536;
static uns string_cats = ~0;
static uns phrase_limit = 4095;
static byte *card_attr_list = "U";
static uns excerpt_min;
static uns excerpt_max = ~0;

static struct cfitem chewer_config[] = {
  { "Chewer",		CT_SECTION,	NULL },
  { "WordBufSize",	CT_INT,		&word_buf_size },
  { "StringBufSize",	CT_INT,		&string_buf_size },
  { "StringCats",	CT_INT,		&string_cats },
  { "PhraseLimit",	CT_INT,		&phrase_limit },
  { "CardAttrs",	CT_STRING,	&card_attr_list },
  { "ExcerptMin",	CT_INT,		&excerpt_min },
  { "ExcerptMax",	CT_INT,		&excerpt_max },
  { NULL,		CT_STOP,	NULL }
};

static void CONSTRUCTOR chewconf_init(void)
{
  cf_register(chewer_config);
}

/*
 *  Indexing of strings
 */

struct sentry {
  struct sentry *next;
  struct fingerprint fp;
  u32 id;
  u32 type;
};

#define SENT_BITE 1024

static struct sentry *string_list;
static uns sentry_count, sentry_limit;
static struct mempool *string_pool;
static uns string_runs;
static u64 string_cnt;

static void
string_init(void)
{
  string_index = bopen(fn_string_index, O_WRONLY | O_CREAT | O_TRUNC, indexer_fb_size);
  string_pool = mp_new(sizeof(struct sentry) * SENT_BITE);
  sentry_limit = string_buf_size / sizeof(struct sentry);
  DBG("Allocated string pool, sentry_limit=%d", sentry_limit);
}

static inline int
string_cmp(const struct sentry *a, const struct sentry *b)
{
  int i = memcmp(&a->fp, &b->fp, sizeof(struct fingerprint));
  if (i)
    return i;
  if (a->id < b->id)
    return -1;
  if (a->id > b->id)
    return 1;
  if (a->type < b->type)
    return -1;
  if (a->type > b->type)
    return 1;
  return 0;
}

static struct sentry *
string_sort(struct sentry *x)
{
  struct sentry *f1, **l1, *f2, **l2, **l;

  l1 = &f1;
  l2 = &f2;
  while (x)
    {
      *l1 = x;
      l1 = &x->next;
      x = x->next;
      if (!x)
	break;
      *l2 = x;
      l2 = &x->next;
      x = x->next;
    }
  *l1 = *l2 = NULL;

  if (f1 && f1->next)
    f1 = string_sort(f1);
  if (f2 && f2->next)
    f2 = string_sort(f2);
  l = &x;
  while (f1 && f2)
    {
      if (string_cmp(f1, f2) <= 0)
	{
	  *l = f1;
	  l = &f1->next;
	  f1 = f1->next;
	}
      else
	{
	  *l = f2;
	  l = &f2->next;
	  f2 = f2->next;
	}
    }
  *l = f1 ? : f2;
  return x;
}

static void
string_flush(void)
{
  uns size;
  struct sentry *f, *g, *h, *m, **hh;
  int prevtype;
  sh_off_t expected_end;

  if (!string_list)
    return;
  string_list = string_sort(string_list);

  f = string_list;
  while (f)
    {
      g = f;
      while (f && !memcmp(&f->fp, &g->fp, sizeof(struct fingerprint)))
	f = f->next;
      /* [g,f) is the chain to flush for single fp, scan sizes */
      size = 0;
      h = g;
      while (h != f)
	{
	  m = h;
	  size += 6;
	  prevtype = -1;
	  hh = NULL;
	  while (h != f && h->id == m->id)
	    {
	      if ((int) h->type == prevtype)
		*hh = h->next;		/* duplicates are deleted */
	      else
		{
		  size += 2;
		  prevtype = h->type;
		  hh = &h->next;
		}
	      h = h->next;
	    }
	}
      /* output the chain */
      bwrite(string_index, &g->fp, sizeof(struct fingerprint));
      bputl(string_index, size);
      expected_end = btell(string_index) + size;
      while (g != f)
	{
	  bputl(string_index, g->id);
	  size = 1;
	  h = g->next;
	  while (h != f && h->id == g->id)
	    {
	      size++;
	      h = h->next;
	    }
	  bputw(string_index, size);
	  h = g;
	  while (g != f && g->id == h->id)
	    {
	      bputw(string_index, (g->type << 12) | 0xfff);
	      g = g->next;
	    }
	}
      ASSERT(btell(string_index) == expected_end);
    }
  string_runs++;
  string_list = NULL;
  sentry_count = 0;
  mp_flush(string_pool);
}

static void
string_end(void)
{
  string_flush();
  bclose(string_index);
  log(L_INFO, "Generated %Ld strings in %d runs", string_cnt, string_runs);
}

static void
string_add(byte *s, uns type)
{
  struct sentry *e;

  if (!(string_cats & (1 << type)))
    return;
  e = mp_alloc(string_pool, sizeof(*e));
  e->next = string_list;
  string_list = e;
  fingerprint(s, &e->fp);
  e->id = card_id;
  e->type = type;
  sentry_count++;
  string_cnt++;
}

static void
string_card(struct odes *o)
{
  struct oattr *a;
  struct { uns attr, type; } strs[] = {
    { 'A', ST_REF },
    { 'F', ST_REF },
    { 'I', ST_REF },
    { 'R', ST_REF },
    { 'Y', ST_REF },
    { 'd', ST_REF },
    { 'f', ST_REF },
    { 'y', ST_URL },
    { 0, 0 }
  };
  uns i;

  for (a=obj_find_attr(o, 'U'); a; a=a->same)
    {
      byte *url = a->val;
      byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE];
      struct url u;
      string_add(url, ST_URL);
      if (!url_canon_split(url, buf1, buf2, &u))
	{
	  byte *dot = strchr(u.host, '.');
	  string_add(u.host, ST_HOST);
	  if (dot && strchr(dot+1, '.'))
	    string_add(dot+1, ST_DOMAIN);
	}
    }

  for (i=0; strs[i].attr; i++)
    for (a=obj_find_attr(o, strs[i].attr); a; a=a->same)
      {
	byte *x = a->val;
	while (*x && *x != ' ')
	  x++;
	if (*x)
	  {
	    *x = 0;
	    string_add(a->val, strs[i].type);
	    *x = ' ';
	  }
	else
	  string_add(a->val, strs[i].type);
      }

  if (sentry_count >= sentry_limit)
    string_flush();
}

/*
 *  Indexing of words
 */

#define LENT_QUANTUM 3
#define LENT_BITE 1024

typedef struct lentry {
  struct lentry *next;
  u32 id;
  u16 count;
  u16 w[LENT_QUANTUM];
} lentry;

static uns lentry_count, lentry_limit;
static struct verbum **used_words, **used_words_last;
static struct mempool *word_pool;
static uns word_runs;
static u64 word_entries, word_cnt;
static uns seen_eox;

#define LH_CHEWER
#include "indexer/lexhash.h"

static void
lex_load(void)
{
  struct fastbuf *b;
  u32 x;
  struct verbum *v;
  uns i;

  b = bopen(fn_lex_temp, O_RDONLY, indexer_fb_size);
  bgetl(b);
  while ((x = bgetl(b)) != ~0UL)
    {
      enum word_class class = bgetc(b);
      uns len = bgetc(b);
      if (class == WC_COMPOUND)
	{
	  u32 buf[MAX_COMPLEX_LEN];
	  bread(b, buf, len);
	  len /= 4;
	  v = ch_insert(buf, len);
	  if (!v)
	    die("Malformed lexicon: Duplicate complex");
	}
      else
	{
	  byte buf[MAX_WORD_LEN+1];
	  breadb(b, buf, len);
	  buf[len] = 0;
	  v = lh_insert(buf, 0);
	  if (!v)
	    die("Malformed lexicon: Duplicate word <%s>", buf);
	  v->id |= class;
	}
      v->u.count = x;
    }
  bclose(b);
  lh_rehash(lh_hash_count);		/* Sort the chains by counts */
  ch_rehash(ch_hash_count);
  for (i=0; i<lh_hash_size; i++)
    for (v=lh_hash_table[i]; v; v=v->next)
      v->u.first_lent = NULL;
  for (i=0; i<ch_hash_size; i++)
    for (v=ch_hash_table[i]; v; v=v->next)
      v->u.first_lent = NULL;
  log(L_INFO, "Read lexicon with %d words and %d complexes", lh_hash_count, ch_hash_count);
}

static void
word_flush_single(uns id, lentry *head)
{
  lentry *f, *g;
  uns size;
  sh_off_t expos;

  /* The chain is reversed, flip it back */
  g = NULL;
  while (head)
    {
      f = head->next;
      head->next = g;
      g = head;
      head = f;
    }
  head = g;

  /* Calculate block size */
  size = 0;
  f = head;
  while (f)
    {
      size += 6;
      g = f;
      while (f && f->id == g->id)
	{
	  size += 2*f->count;
	  f = f->next;
	}
    }

  /* Write header */
  bputl(word_index, id);
  bputl(word_index, size);
  expos = btell(word_index) + size;

  /* Write body */
  f = head;
  while (f)
    {
      bputl(word_index, f->id);
      g = f;
      size = 0;
      while (g && g->id == f->id)
	{
	  size += g->count;
	  g = g->next;
	}
      bputw(word_index, size);
      g = f;
      while (f && g->id == f->id)
	{
	  bwrite(word_index, f->w, 2*f->count);
	  f = f->next;
	}
    }

  /* Sanity check */
  if (expos != btell(word_index))
    die("word_flush_single: Internal error, size mismatch");
}

static void
word_sort(void)
{
  /*
   *  We don't use qsort(), because we want to avoid the overhead
   *  of callbacks. Instead, we do a simple QuickSort borrowed from
   *  expire.c. One day, somebody should make a generic library macro
   *  for such sorting.
   */
  struct verbum **base = used_words;
  struct stk { int l, r; } stack[32];
  int l, r, left, right, m;
  uns sp = 0;
  u32 pivot;
  struct verbum *tmp;
#define SWAP(x,y) do { tmp=base[x]; base[x]=base[y]; base[y]=tmp; } while(0)

  left = 0;
  right = used_words_last - used_words - 1;
  if (left >= right)
    return;
  for(;;)
    {
      l = left;
      r = right;
      m = (l+r)/2;
      if (base[l]->id > base[m]->id)
	SWAP(l,m);
      if (base[m]->id > base[r]->id)
	{
	  SWAP(m,r);
	  if (base[l]->id > base[m]->id)
	    SWAP(l,m);
	}
      pivot = base[m]->id;
      do
	{
	  while (base[l]->id < pivot)
	    l++;
	  while (base[r]->id > pivot)
	    r--;
	  if (l < r)
	    {
	      SWAP(l,r);
	      l++;
	      r--;
	    }
	  else if (l == r)
	    {
	      l++;
	      r--;
	    }
	}
      while (l <= r);
      if (left < r && l < right)
	{
	  /* Both partitions ok => push the larger one */
	  if ((r - left) > (right - l))
	    {
	      stack[sp].l = left;
	      stack[sp].r = r;
	      left = l;
	    }
	  else
	    {
	      stack[sp].l = l;
	      stack[sp].r = right;
	      right = r;
	    }
	  sp++;
	}
      else if (left < r)
	{
	  /* Left partition OK, right undersize */
	  right = r;
	}
      else if (l < right)
	{
	  /* Right partition OK, left undersize */
	  left = l;
	}
      else
	{
	  /* Both partitions undersize => pop */
	  if (!sp)
	    break;
	  sp--;
	  left = stack[sp].l;
	  right = stack[sp].r;
	}
    }
#undef SWAP
}

static void
word_flush(void)
{
  struct verbum **w, *v;

  word_sort();
  *used_words_last = NULL;
  for (w=used_words; v = *w; w++)
    {
      word_flush_single(v->id/8, v->u.first_lent);
      v->u.first_lent = NULL;
    }

  mp_flush(word_pool);
  word_runs++;
  word_entries += lentry_count;
  lentry_count = 0;
  used_words_last = used_words;
}

static inline int
word_check(struct verbum *v, u16 pack)
{
  lentry *e = v->u.first_lent;

  while (e && e->id == card_id)
    {
      int i = e->count - 1;
      while (i >= 0)
	{
	  if (e->w[i] == pack)
	    return 1;
	  if ((e->w[i] ^ pack) & 0xfff)
	    return 0;
	  i--;
	}
      e = e->next;
    }
  return 0;
}

static inline void
word_add(struct verbum *v, u16 pack)
{
  lentry *e = v->u.first_lent;
  if (!e || e->id != card_id || e->count >= LENT_QUANTUM)
    {
      e = mp_alloc(word_pool, sizeof(lentry));
      if (!v->u.first_lent)
	*used_words_last++ = v;
      e->next = v->u.first_lent;
      v->u.first_lent = e;
      e->id = card_id;
      e->count = 0;
      lentry_count++;
    }
  e->w[e->count++] = pack;
  word_cnt++;
}

static enum word_class
lm_lookup(enum word_class orig_class, word *uni, uns ulen, word_id_t *idp)
{
  struct verbum *v;

  if (orig_class != WC_NORMAL)
    return orig_class;
  v = lh_lookup(uni, ulen);
  *idp = v;
  return v->id & 7;
}

static void
lm_got_word(uns pos, uns cat, word_id_t w)
{
  if (pos >= phrase_limit)
    {
      if (!word_check(w, (cat << 12) | 4095))
	word_add(w, (cat << 12) | 4095);
      seen_eox = 1;
    }
  else
    word_add(w, (cat << 12) | pos);
}

static inline void
lm_got_compound(uns pos, uns cat, word_id_t *w, uns l)
{
  struct verbum *v;

  v = ch_lookup(w, l);
  lm_got_word(pos, cat, v);
}

#include "indexer/lexmap.h"

static void
word_card(struct odes *o)
{
  struct oattr *a;

  lm_doc_start();
  descr_stop = NULL;
  seen_eox = 0;
  for (a=obj_find_attr(o, 'X'); a; a=a->same)
    {
      uns len = str_len(a->val);
      if (seen_eox && !descr_stop)
	descr_stop = a->val;
      lm_map_text(a->val, a->val + len);
    }

  if (lentry_count >= lentry_limit)
    word_flush();
}

static void
word_init(void)
{
  lh_init();
  lm_init();
  word_index = bopen(fn_word_index, O_WRONLY | O_CREAT | O_TRUNC, indexer_fb_size);
  lex_load();
  word_pool = mp_new(sizeof(lentry) * LENT_BITE);
  lentry_limit = word_buf_size / (sizeof(lentry) + sizeof(struct lentry *));
  /*
   * The used_words array is large, but most of it will be probably left unmapped.
   * Unfortunately, we don't have a better bound since lentry_count >= lentry_limit
   * testing is deffered to card end (which is faster and it also guarantees that no
   * ref chain for a (word,card) pair will be split, which is required by wsort).
   */
  used_words_last = used_words = xmalloc(sizeof(struct lentry *) * (lh_hash_count + ch_hash_count + 1));
  DBG("Allocated word pool, lentry_limit=%d", lentry_limit);
}

static void
word_end(void)
{
  word_flush();
  bclose(word_index);
  log(L_INFO, "Generated %Ld word refs in %d runs", word_cnt, word_runs);
  DBG("Used %Ld entries, that is %Ld bytes", word_entries, word_entries * sizeof(lentry));
}

/*
 *  Processing of cards
 */

static void
cards_init(void)
{
  struct card_attr a;

  cards = bopen(fn_cards, O_WRONLY | O_CREAT | O_TRUNC, indexer_fb_size);
  card_attrs = bopen(fn_card_attrs, O_WRONLY | O_CREAT | O_TRUNC, indexer_fb_size);
  bzero(&a, sizeof(a));			/* Create dummy document 0 to make all read ID's be >0 */
  bwrite(card_attrs, &a, sizeof(a));
  card_id++;
}

static void
card_write_start(struct card_attr *attr)
{
  uns align = (1 << CARD_POS_SHIFT) - 1;
  sh_off_t pos;

  bputc(cards, 0);
  pos = btell(cards);
  while (pos & align)
    {
      bputc(cards, 0);
      pos++;
    }
  if ((u64)(pos >> CARD_POS_SHIFT) >= 0xffffffff)
    die("Card file too large. You need to increase CARD_POS_SHIFT in lib/index.h.");
  attr->card = pos >> CARD_POS_SHIFT;
  bwrite(card_attrs, attr, sizeof(*attr));
}

static void
cards_end(void)
{
  struct card_attr a;

  bzero(&a, sizeof(a));			/* Append fake attribute marking end of card file */
  card_write_start(&a);
  bclose(cards);
  bclose(card_attrs);
  log(L_INFO, "Generated %d cards", card_id);
}

static void
init_card(uns id UNUSED, struct card_attr *attr)
{
  card_write_start(attr);
}

static void
chew_line(byte *line)
{
  if (strchr(card_attr_list, line[0]))
    bputsn(cards, line);
}

static void
card_card(struct odes *o)
{
  struct oattr *at;
  uns size, c;
  uns seen_stop = 0;
  sh_off_t start = btell(cards);

  /* We use descr_stop as set by word_card() */
  bputc(cards, 'X');
  for (at=obj_find_attr(o, 'X'); at; at=at->same)
    {
      byte *a = at->val;
      if (a == descr_stop)
	seen_stop = 1;
      size = btell(cards) - start;
      if (size > excerpt_max || seen_stop && size > excerpt_min)
	break;
      if (*a < 0x80 || *a >= 0xa0)
	bputc(cards, ' ');
      while (c = *a++)
	{
	  if (c < 0xa0)
	    bputc(cards, c);
	  else if (c < 0xc0)
	    {
	      /* Brackets are to be removed */
	      if (c < 0xb0 && *a)
		a++;
	    }
	  else
	    {
	      bputc(cards, c);
	      while (*a && (c & 0x40))
		{
		  bputc(cards, *a++);
		  c <<= 1;
		}
	    }
	}
    }
  bputc(cards, '\n');
}

/*
 *  Main loop
 */

static void
chew_card(struct card_attr *attr UNUSED, struct odes *o)
{
  prof_switch(&pr_fetch, &pr_strings);
  string_card(o);
  prof_switch(&pr_strings, &pr_words);
  word_card(o);
  prof_switch(&pr_words, &pr_cards);
  card_card(o);
  prof_switch(&pr_cards, &pr_fetch);
  card_id++;
  if (progress && !(card_id % progress))
    {
      printf("chewer: %d cards\r", card_id);
      fflush(stdout);
    }
}

int
main(int argc, char **argv)
{
  log_init(argv[0]);
  if (cf_getopt(argc, argv, CF_SHORT_OPTS, CF_NO_LONG_OPTS, NULL) >= 0 ||
      optind < argc)
  {
    fputs("This program supports only the following command-line arguments:\n" CF_USAGE, stderr);
    exit(1);
  }

  cards_init();
  word_init();
  string_init();

  prof_init(&pr_fetch);
  prof_init(&pr_strings);
  prof_init(&pr_words);
  prof_init(&pr_cards);

  prof_start(&pr_fetch);
  log(L_INFO, "Chewing cards and creating indices");
  fetch_cards(init_card, chew_line, chew_card);
  prof_stop(&pr_fetch);

  cards_end();
  word_end();
  string_end();

#ifdef PROFILER
  log(L_DEBUG, "Profile: fetch %s, strings %s, words %s, cards %s", PROF_STR(pr_fetch),
      PROF_STR(pr_strings), PROF_STR(pr_words), PROF_STR(pr_cards));
#endif

  return 0;
}
