/*
 *	Sherlock Search Engine -- Query Parser
 *
 *	(c) 1997--2002 Martin Mares <mj@ucw.cz>
 */

%{
#include "lib/lib.h"
#include "lib/pools.h"
#include "search/sherlockd.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>

#define yyerror(x) err(x)

static byte *error;
static int yyparse(void);
static void parse_custom(struct query *q, uns id, uns op, byte *str, uns num, struct val_set *set);

struct expr *new_node(enum expr_type t)
{
  struct expr *n = mp_alloc_zero(current_query->pool, sizeof(struct expr));
  n->type = t;
  return n;
}

struct expr *new_op(enum expr_type t, struct expr *l, struct expr *r)
{
  struct expr *n = mp_alloc_zero(current_query->pool, sizeof(struct expr));
  n->type = t;
  n->u.op.l = l;
  n->u.op.r = r;
  return n;
}

%}

%union {
  int i;
  struct query *q;
  struct expr *x;
  struct val_set *t;
  byte *s;
}

/* Tokens */

%token EOLN K_DEBUG CONTROL ACCENTS LIST SHOW CONTEXT TITLELEN DB SITE URLS SITEMAX
%token PARTIAL APPROX ANY DOTDOT STATS AUTONEAR LE GE NE SORTBY
%token <i> REF NUM SITEID CUSTOM
%token <s> STRING

%left OR
%left AND
%left NOT MAYBE

%type <x> expr atom radical simple molecule neg_molecule override overrule
%type <i> refmask refmask_list maybemin maybemax db_set db_num custom_op
%type <i> maybe_minus custom_attr
%type <q> global querystart selector
%type <t> int_set set setnode

%%

main:
   global expr EOLN {
      $1->expr = $2;
      return 0;
   }
 | global CONTROL STRING {
      $1->cmd = $3;
      return 0;
   }
 ;

querystart:
   /* empty */ { $$ = current_query; }
 ;

selector:
   querystart
 | querystart LIST int_set {
      $1->range = $3;
      $1->list_only = 1;
      $$ = $1;
   }
 | querystart SHOW int_set {
      $1->range = $3;
      $1->list_only = 0;
      $$ = $1;
   }
 | querystart STATS {
      $1->list_only = 2;
      $$ = $1;
   }
 ;

global:
   selector
 | global K_DEBUG NUM {
      $1->debug = $3;
      $$ = $1;
   }
 | global CONTEXT NUM {
      $1->context_chars = $3;
      $$ = $1;
   }
 | global TITLELEN NUM {
      $1->title_chars = $3;
      $$ = $1;
   }
 | global ACCENTS NUM {
      if ($3 > 2)
        err("ACCENTS out of range");
      $1->default_options.accent_mode = $3;
      $$ = $1;
   }
 | global DB db_set {
      $1->db_mask = $3;
      $$ = $1;
   }
 | global SITE SITEID {
      $1->site_only = $3;
      $$ = $1;
   }
 | global SITEMAX NUM {
      if ($3 > 2)
	err("SITEMAX out of range");
      $1->site_max = $3;
      $$ = $1;
   }
 | global URLS NUM {
      $1->url_max = $3;
      $$ = $1;
   }
 | global PARTIAL NUM {
      $1->partial_answers = $3;
      $$ = $1;
   }
 | global APPROX NUM {
      $1->allow_approx = $3;
      $$ = $1;
   }
 | global CUSTOM custom_op NUM {
      parse_custom($1, $2, $3, NULL, $4, NULL);
      $$ = $1;
   }
 | global CUSTOM custom_op STRING {
      parse_custom($1, $2, $3, $4, 0, NULL);
      $$ = $1;
   }
 | global CUSTOM custom_op '{' set '}' {
      parse_custom($1, $2, $3, NULL, 0, $5);
      $$ = $1;
   }
 | global SORTBY maybe_minus custom_attr {
      $1->custom_sorting = $4;
      $1->custom_sort_reverse = ($3 ? ~0U : 0);
      $$ = $1;
   }
 ;

custom_op:
   '<' { $$ = 0; }
 | '>' { $$ = 1; }
 | LE  { $$ = 2; }
 | GE  { $$ = 3; }
 | '=' { $$ = 4; }
 | NE  { $$ = 5; }
 ;

custom_attr:
   CUSTOM
 | SITE { $$ = PARAM_SITE; }
 ;

maybe_minus:
   '-' { $$ = 1; }
 | /* empty */ { $$ = 0; }
 ;

expr:
   molecule
 | expr AND expr {
     $$ = new_op(EX_AND, $1, $3);
   }
 | expr OR expr {
     $$ = new_op(EX_OR, $1, $3);
   }
 ;

molecule:
   neg_molecule
 | simple
 ;

neg_molecule:
   '(' expr ')' override {
     if ($4)
       {
	 $4->u.options.inside = $2;
	 $$ = $4;
       }
     else
       $$ = $2;
   }
 | ANY { $$ = new_node(EX_ANY); }
 | NOT neg_molecule {
     $$ = new_op(EX_NOT, $2, NULL);
   }
 ;

atom:
   radical       { $$=$1; $$->u.match.sense=1; }
 | NOT radical   { $$=$2; $$->u.match.sense=-1; }
 | MAYBE radical { $$=$2; $$->u.match.sense=0; }
 ;

radical:
   refmask STRING override {
     $$ = new_node(EX_MATCH);
     $$->u.match.classmap = $1 & 0xffff;
     $$->u.match.is_string = !!($1 & 0x40000000);
     $$->u.match.word = $2;
     if ($3)
       $$->u.match.o = $3->u.options.o;
     else
       {
	 $$->u.match.o.accent_mode = ACCENT_DEFAULT;
	 $$->u.match.o.weight = WEIGHT_DEFAULT;
       }
     $$->u.match.next_simple = NULL;
   }
 ;

simple:
   atom
 | atom '.' simple { $$=$1; $1->u.match.next_simple = $3; }
 ;

refmask:
   /* empty */ { $$ = default_word_types; }
 | refmask_list
 ;

refmask_list:
   REF { $$ = $1; }
 | refmask_list ',' REF {
     if (($1 ^ $3) & 0x40000000)
       err("Incompatible word types");
     $$ = $1 | $3;
   }
 ;

override:
   /* empty */ { $$ = NULL; }
 | overrule '/' NUM {
     if ($3 > 30000)
       err("Word weight out of range");
     $$ = $1;
     $$->u.options.o.weight = $3;
   }
 | overrule '/' '-' NUM {
     if ($4 > 30000)
       err("Word weight out of range");
     $$ = $1;
     $$->u.options.o.weight = -$4;
   }
 | overrule ACCENTS NUM {
     if ($3 > 2)
       err("ACCENTS out of range");
     $$ = $1;
     $$->u.options.o.accent_mode = $3;
   }
 ;

overrule:
   override {
     $$ = $1;
     if (!$$)
       {
	 $$ = new_node(EX_OPTIONS);
	 $$->u.options.o.weight = WEIGHT_DEFAULT;
	 $$->u.options.o.accent_mode = ACCENT_DEFAULT;
       }
   }
 ;

int_set:
   set {
     struct val_set *t;
     for (t=$1; t; t=t->next)
       if (t->text)
	 err("Integer set expected");
     $$ = $1;
   }
 ;

set:
   setnode { $1->next = NULL; $$ = $1; }
 | set ',' setnode { $3->next = $1; $$ = $3; }
 ;

setnode:
   NUM {
     $$ = mp_alloc(current_query->pool, sizeof(struct val_set));
     $$->min = $$->max = $1;
     $$->text = NULL;
   }
 | STRING {
     $$ = mp_alloc(current_query->pool, sizeof(struct val_set));
     $$->min = $$->max = 0;
     $$->text = $1;
   }
 | maybemin DOTDOT maybemax {
     $$ = mp_alloc(current_query->pool, sizeof(struct val_set));
     if ($1 > $3)
       err("Invalid interval");
     $$->min = $1;
     $$->max = $3;
     $$->text = NULL;
   }
 ;

maybemin:
   /* empty */ { $$ = 0; }
 | NUM { $$ = $1; }
 ;

maybemax:
   /* empty */ { $$ = ~0; }
 | NUM { $$ = $1; }
 ;

db_set:
   db_num { $$ = $1; }
 | db_set ',' db_num { $$ = $1 | $3; }
 ;

db_num:
   STRING {
     struct database *db;
     int i;
     for (db=databases, i=0; db; db=db->next, i++)
       if (!strcmp(db->name, $1))
         break;
     if (!db)
       err("Unknown database");
     $$ = 1 << i;
   }
 ;

%%

static jmp_buf err_jmp;

void
err(byte *x)
{
  error = x;
  longjmp(err_jmp, 1);
}

byte *
parse_query(byte *b)
{
  error = NULL;
  lex_init(b);
  if (!setjmp(err_jmp))
    yyparse();
  return error;
}

static byte *
parse_age(u32 *dest, byte *val, uns intval)
{
  if (val)
    return "Incorrect AGE";
  *dest = intval;
  return NULL;
}

static void
parse_custom(struct query *q, uns id, uns op, byte *str, uns num, struct val_set *set)
{
  byte *(*pfunc)(u32 *dest, byte *value, uns intval);
  byte *msg;
  u32 val;
  u32 min = 0, max = ~0U, empty = 0;
  u32 *pmin, *pmax, *pset;

  switch (id)
    {
    case PARAM_AGE:
      pfunc = parse_age;
      pmin = &q->age_raw_min;
      pmax = &q->age_raw_max;
      pset = NULL;
      break;
#define INT_ATTR(id,keywd,gf,pf)		\
    case OFFSETOF(struct query, id##_min):	\
      pfunc = pf;				\
      pmin = &q->id##_min;			\
      pmax = &q->id##_max;			\
      pset = NULL;				\
      break;
#define SMALL_SET_ATTR(id,keywd,gf,pf) 		\
    case OFFSETOF(struct query, id##_set):	\
      pfunc = pf;				\
      pmin = pmax = NULL;			\
      pset = &q->id##_set;			\
      break;
    CUSTOM_ATTRS
#undef INT_ATTR
#undef SMALL_SET_ATTR
    default:
      die("parse_custom: unknown attribute");
    }

  if (set)
    {
      if (!pset)
	err("Set matching not supported for this attribute");
      if (op != 4 && op != 5)
	err("Sets are not ordered");
      *pset = 0;
      while (set)
	{
	  if (set->text)
	    {
	      u32 val;
	      if (msg = pfunc(&val, set->text, 0))
		err(msg);
	      *pset |= 1 << val;
	    }
	  else
	    {
	      u32 min, max;
	      if ((msg = pfunc(&min, NULL, set->min)) || (msg = pfunc(&max, NULL, set->max)))
		err(msg);
	      while (min <= max)
		*pset |= 1 << min++;
	    }
	  set = set->next;
	}
      if (op == 5)
        *pset = ~*pset;
      return;
    }

  if (msg = pfunc(&val, str, num))
    err(msg);

  switch (op)
    {
    case 0:
      /* Unfortunately, we must handle <0 separately as we don't use signed integers */
      if (val) max=val-1; else empty=1;
      break;
    case 1:
      if (val == ~0U) empty=1; else min=val+1;
      break;
    case 2:
      max=val;
      break;
    case 3:
      min=val;
      break;
    case 4:
      min=max=val;
      break;
    case 5:				/* <> supported only for set attributes */
      if (!pset)
	err("<> supported only for set-matched attributes");
      *pset &= ~(1 << val);
      return;
    default:
      ASSERT(0);
    }
  if (pset)
    {
      u32 mask = 0;
      while (min <= MIN(max, 31))
	mask |= 1 << min++;
      *pset &= mask;
    }
  else
    {
      *pmin = MAX(*pmin, min);
      *pmax = MIN(*pmax, max);
      if (*pmin > *pmax || empty)
	{
	  *pmin = ~0U;
	  *pmax = 0;
	}
    }
}
