/*
 *	Sherlock Search Engine -- Main
 *
 *	(c) 1997--2002 Martin Mares <mj@ucw.cz>
 */

#include "lib/lib.h"
#include "lib/conf.h"
#include "lib/pools.h"
#include "lib/ipaccess.h"
#include "search/sherlockd.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

/*** Allocation and freeing of queries ***/

static list query_list;

static struct query *
new_query(void)
{
  struct mempool *pool = mp_new(8192);
  struct query *q = mp_alloc_zero(pool, sizeof(struct query));

  q->pool = pool;
  init_reply_buf(&q->rbuf, pool);
  q->ibptr = q->iobuf;
  q->ibend = q->iobuf + sizeof(q->iobuf);
  q->obptr = q->iobuf;
  q->obend = q->iobuf + sizeof(q->iobuf);
  add_tail(&query_list, &q->n);
  memory_setup(q);
  q->q_status = -1;
  return q;
}

static void
free_query(struct query *q)
{
  rem_node(&q->n);
  memory_flush(q);
  mp_delete(q->pool);
}

/*** Recoding and I/O ***/

static void
flush_buffer(struct query *q)
{
  int l, e;
  byte *buf = q->iobuf;
  prof_t *old_prof;

  old_prof = profiler_switch(&prof_send);
  l = q->obptr - q->iobuf;
  while (l && !q->fd_err)
    {
      e = write(q->fd, buf, l);
      if (e <= 0)
	q->fd_err = 1;
      else
	{
	  buf += e;
	  l -= e;
	}
    }
  q->obptr = q->iobuf;
  profiler_switch(old_prof);
}

void
reply_string(struct query *q, byte *data, uns len)
{
  uns r;

  while (len)
    {
      r = q->obend - q->obptr;
      if (r > len)
	r = len;
      if (!r)
	{
	  flush_buffer(q);
	  continue;
	}
      memcpy(q->obptr, data, r);
      q->obptr += r;
      data += r;
      len -= r;
    }
}

/*** Handling of connections ***/

static void
close_conn(struct query *q, byte *err)
{
  if (err)
    {
      if (log_replies)
	log(L_INFO, "%s : %s", q->ipaddr, err);
      add_reply(&q->rbuf, err);
    }
  flush_reply_buf(q, &q->rbuf);
  reply_string(q, "\n", 1);
  flush_buffer(q);
  close(q->fd);
  free_query(q);
}

static void
incoming_conn(int sock)
{
  int sock2, alen;
  struct sockaddr_in addr;
  u32 ip;
  struct query *q;

  alen = sizeof(addr);
  sock2 = accept(sock, (struct sockaddr *) &addr, &alen);
  if (sock2 < 0)
    {
      if (errno != EAGAIN)
	log(L_ERROR, "accept(): %m");
      return;
    }
  if (fcntl(sock2, F_SETFL, O_NONBLOCK))
    {
      log(L_ERROR, "fcntl(): %m");
      return;
    }

  q = new_query();
  q->fd = sock2;
  q->established = time(NULL);

  ip = ntohl(addr.sin_addr.s_addr);
  sprintf(q->ipaddr, "%d.%d.%d.%d", (ip>>24)&255, (ip>>16)&255, (ip>>8)&255, ip&255);
  if (!ipaccess_check(access_list, ip))
    {
      if (log_rejected)
	log(L_INFO, "Rejected connection from %s", q->ipaddr);
      close_conn(q, "-100 Apage Satanas!");
      return;
    }

  if (log_incoming)
    log(L_INFO, "Accepted connection from %s, fd=%d", q->ipaddr, sock2);
}

static void
continue_conn(struct query *q)
{
  int sz, mx;
  byte *r;

  mx = q->ibend - q->ibptr - 1;
  sz = read(q->fd, q->ibptr, mx);
  DBG("Data from %d: %d bytes", q->fd, sz);
  if (!sz)
    {
      close_conn(q, "-108 Incomplete request");
      return;
    }
  if (sz < 0)
    {
      if (errno == EAGAIN)
	return;
      close_conn(q, "-108 Read error");
      return;
    }
  if (sz >= mx)
    {
      close_conn(q, "-101 Request too long");
      return;
    }
  if (fcntl(q->fd, F_SETFL, 0))		/* writes will be blocking */
    die("fcntl(): %m");
  r = q->ibptr;
  while (sz--)
    {
      if (*r == '\r' || *r == '\n')
	{
	  *r = 0;
	  if (query_watchdog)
	    alarm(query_watchdog);
	  if (log_requests)
	    log(L_INFO, "%s < %s", q->ipaddr, q->iobuf);
	  process_query(q);
	  if (query_watchdog)
	    alarm(0);
	  if (log_replies)
	    {
	      byte sprof[PROF_STR_SIZE+6];
#ifdef PROFILER
	      strcpy(sprof, " send=");
	      prof_format(sprof+6, &prof_send);
#else
	      sprof[0] = 0;
#endif
	      log(L_INFO, "> %d t=%d%s%s%s", q->q_status, q->time_total,
		  q->profile_stats ? " " : "",
		  q->profile_stats ? : "",
		  sprof);
	    }
	  close_conn(q, NULL);
	  return;
	}
      r++;
    }
  q->ibptr = r;
}

static void
watchdog_timeout(int x UNUSED)
{
  die("Watchdog timeout");
}

static void
mainloop(void)
{
  int sock, n;
  int one = 1;
  time_t now;
  fd_set fds;
  struct query *q, *nq;
  struct timeval tv;
  struct sockaddr_in addr;
  uns timeout;

  signal(SIGPIPE, SIG_IGN);
  signal(SIGALRM, watchdog_timeout);
  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0)
    die("socket(): %m");
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
    die("setsockopt(): %m");
  if (fcntl(sock, F_SETFL, O_NONBLOCK))
    die("fcntl(): %m");
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)))
    die("bind(): %m");
  if (listen(sock, listen_queue))
    die("listen(): %m");
  log(L_INFO, "Listening at port %d", port);

  init_list(&query_list);

  for(;;)
    {
      now = time(NULL);
      n = sock;
      FD_ZERO(&fds);
      FD_SET(sock, &fds);
      timeout = 1000000;
      for(q = HEAD(query_list); nq = (void *) q->n.next; q = nq)
	{
	  if (now - q->established >= connection_timeout)
	    close_conn(q, "-107 Timed out");
	  else
	    {
	      uns rem = q->established + connection_timeout - now;
	      if (rem < timeout)
		timeout = rem;
	      FD_SET(q->fd, &fds);
	      if (q->fd > n)
		n = q->fd;
	    }
	}

      tv.tv_sec = timeout;
      tv.tv_usec = 0;
      n = select(n+1, &fds, NULL, NULL, &tv);
      if (n < 0)
	die("Select failed: %m");
      if (!n)
	continue;

      if (FD_ISSET(sock, &fds))
	incoming_conn(sock);

      for(q = HEAD(query_list); nq = (void *) q->n.next; q = nq)
	if (FD_ISSET(q->fd, &fds))
	  continue_conn(q);
    }
}

/*** Main ***/

int
main(int argc, char **argv)
{
  log_init(NULL);
  if (cf_getopt(argc, argv, CF_SHORT_OPTS, CF_NO_LONG_OPTS, NULL) >= 0 ||
      optind < argc)
  {
    fputs("This program supports only the following command-line arguments:\n" CF_USAGE, stderr);
    exit(1);
  }
  log_file(log_name);

  db_init();
  cache_init();
  query_init();
  refs_init();
  memory_init();
  mainloop();
  return 0;
}
